#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier
import librosa
from spafe.features.gfcc import gfcc

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
# def train_challenge_model(data_folder, model_folder, verbose):
#     # Find data files.
#     if verbose >= 1:
#         print('Finding data files...')
#
#
#     # Find the patient data files.
#     patient_files = find_patient_files(data_folder)
#     num_patient_files = len(patient_files)
#
#     if num_patient_files==0:
#         raise Exception('No data was provided.')
#
#     # Create a folder for the model if it does not already exist.
#     os.makedirs(model_folder, exist_ok=True)
#
#     classes = ['Present', 'Unknown', 'Absent']
#     num_classes = len(classes)
#
#     # Extract the features and labels.
#     if verbose >= 1:
#         print('Extracting features and labels from the Challenge data...')
#
#     features = list()
#     labels = list()
#
#     for i in range(num_patient_files):
#         if verbose >= 2:
#             print('    {}/{}...'.format(i+1, num_patient_files))
#
#         # Load the current patient data and recordings.
#         current_patient_data = load_patient_data(patient_files[i]) # read txt handle
#
#         current_recordings = load_recordings(data_folder, current_patient_data)
#
#         # Extract features.
#         current_features = get_features(current_patient_data, current_recordings)
#         features.append(current_features)
#
#         # Extract labels and use one-hot encoding.
#         current_labels = np.zeros(num_classes, dtype=int)
#         label = get_murmur(current_patient_data)
#         if label in classes:
#             j = classes.index(label)
#             current_labels[j] = 1
#         labels.append(current_labels)
#
#     features = np.vstack(features)
#     labels = np.vstack(labels)
#
#     # Train the model.
#     if verbose >= 1:
#         print('Training model...')
#
#     # Define parameters for random forest classifier.
#     n_estimators = 10    # Number of trees in the forest.
#     max_leaf_nodes = 100 # Maximum number of leaf nodes in each tree.
#     random_state = 123   # Random state; set for reproducibility.
#
#     imputer = SimpleImputer().fit(features)
#     features = imputer.transform(features)
#     classifier = RandomForestClassifier(n_estimators=n_estimators,
#         max_leaf_nodes=max_leaf_nodes,
#         random_state=random_state).fit(features, labels)
#
#     # Save the model.
#     save_challenge_model(model_folder, classes, imputer, classifier)
#
#     if verbose >= 1:
#         print('Done.')
#

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')


    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    classes = ['Present', 'Unknown', 'Absent']
    num_classes = len(classes)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    features = list()
    labels = list()

    for i in range(num_patient_files):
        if verbose >= 2:
            print('    {}/{}...'.format(i+1, num_patient_files))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i]) # read txt handle

        current_recordings = load_recordings(data_folder, current_patient_data)

        recording_loc = checkRecordingsName(data_folder, current_patient_data)

        # for each recording, predict its class: [murmur_present, murmur_absent, murmur_unknown]
        for record_idx, record in enumerate(current_recordings):
            # Extract features.
            record_feature = get_each_feature(current_patient_data, record)

            features.append(record_feature)

            # Extract labels and use one-hot encoding.
            current_labels = np.zeros(num_classes, dtype=int)
            present, location = checkPresentAudible(current_patient_data) # get GT murmur location
            cur_audio_loc = recording_loc[record_idx] # get current audio location name
            if cur_audio_loc in location and present == 'Present': # this audio has murmur
                current_labels[0] = 1
            elif cur_audio_loc not in location and present == 'Present':
                current_labels[2] = 1
            elif present == 'Absent':
                current_labels[2] = 1
            elif present == 'Unknown':
                current_labels[1] = 1
            labels.append(current_labels)

    features = np.vstack(features)
    labels = np.vstack(labels)

    # Train the model.
    if verbose >= 1:
        print('Training model...')

    # Define parameters for random forest classifier.
    n_estimators = 10    # Number of trees in the forest.
    max_leaf_nodes = 100 # Maximum number of leaf nodes in each tree.
    random_state = 123   # Random state; set for reproducibility.

    imputer = SimpleImputer().fit(features)
    features = imputer.transform(features)
    classifier = RandomForestClassifier(n_estimators=n_estimators,
        max_leaf_nodes=max_leaf_nodes,
        random_state=random_state,
        class_weight="balanced").fit(features, labels)

    # Save the model.
    save_challenge_model(model_folder, classes, imputer, classifier)

    if verbose >= 1:
        print('Done.')



def checkPresentAudible(txt_file_handle):
    location = 'nan'
    present = 'nan'

    for i, l in enumerate(txt_file_handle.split('\n')):
        if '#Murmur locations:' in l:
            try:
                location = (l.split(':')[1])
                location = (location.split(' ')[1])
            except:
                pass
        elif '#Murmur:' in l:
            try:
                present = (l.split(':')[1])
                present = (present.split(' ')[1])
            except:
                pass

    return present, location


def checkRecordingsName(data_folder, data, get_frequencies=False):
    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations+1]

    recordings_name = list()
    for i in range(num_locations):
        entries = recording_information[i].split(' ')
        recording_file = entries[0] # 'av','pv','tv','mv',...

        recordings_name.append(recording_file)

    return recordings_name



# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    filename = os.path.join(model_folder, 'model.sav')
    return joblib.load(filename)

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
# def run_challenge_model(model, data, recordings, verbose):
#     classes = model['classes']
#     imputer = model['imputer']
#     classifier = model['classifier']
#
#     # Load features.
#     features = get_features(data, recordings)
#
#     # Impute missing data.
#     features = features.reshape(1, -1)
#     features = imputer.transform(features)
#
#     # Get classifier probabilities.
#     probabilities = classifier.predict_proba(features)
#     probabilities = np.asarray(probabilities, dtype=np.float32)[:, 0, 1]
#
#     # Choose label with higher probability.
#     labels = np.zeros(len(classes), dtype=np.int_)
#     idx = np.argmax(probabilities)
#     labels[idx] = 1
#
#     return classes, labels, probabilities

def run_challenge_model(model, data, recordings, verbose):
    classes = model['classes']
    imputer = model['imputer']
    classifier = model['classifier']

    final_probability = np.zeros(3, dtype=np.float32)

    for record_idx, record in enumerate(recordings):

        # Load features.
        features = get_each_feature(data, record)

        # Impute missing data.
        features = features.reshape(1, -1)
        features = imputer.transform(features)

        # Get classifier probabilities.
        probabilities = classifier.predict_proba(features)
        probabilities = np.asarray(probabilities, dtype=np.float32)[:, 0, 1]

        final_probability += probabilities

    final_probability /= float(len(recordings))
    # Choose label with higher probability.
    labels = np.zeros(len(classes), dtype=np.int_)
    idx = np.argmax(final_probability)
    labels[idx] = 1

    return classes, labels, final_probability


################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
def save_challenge_model(model_folder, classes, imputer, classifier):
    d = {'classes': classes, 'imputer': imputer, 'classifier': classifier}
    filename = os.path.join(model_folder, 'model.sav')
    joblib.dump(d, filename, protocol=0)

# Extract features from the data.
def get_features(data, recordings):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    # Extract recording locations and data. Identify when a location is present,
    # and compute the mean, variance, and skewness of each recording. If there
    # are multiple recordings for one location, then extract features from the
    # last recording.
    locations = get_locations(data)

    # recording_locations = ['AV', 'MV', 'PV', 'TV', 'PhC']
    # num_recording_locations = len(recording_locations)

    num_mfcc_features = 10
    # recording_features = np.zeros((num_recording_locations, 4 + num_mfcc_features * 3), dtype=float)
    recording_features = np.zeros((len(locations), 4 + num_mfcc_features * 3), dtype=float)
    num_locations = len(locations)

    num_recordings = len(recordings)
    if num_locations==num_recordings:
        for i in range(num_locations):
            # for j in range(num_recording_locations):
                # if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i])>0:
            if np.size(recordings[i])>0:
                # compute mfcc feature
                mfcc = librosa.feature.mfcc(y=recordings[i].astype('float'), n_mfcc=num_mfcc_features)
                mfcc -= (np.mean(mfcc, axis=0) + 1e-8) # normalization

                mfcc_means = mfcc.mean(axis=1)
                mfcc_vars = mfcc.var(axis=1)
                mfcc_skew = sp.stats.skew(mfcc, axis=1)

                recording_features[j, 0] = 1
                recording_features[j, 1] = np.mean(recordings[i])
                recording_features[j, 2] = np.var(recordings[i])
                recording_features[j, 3] = sp.stats.skew(recordings[i])
                recording_features[j, 4:4+num_mfcc_features] = mfcc_means
                recording_features[j, 4+num_mfcc_features:4+num_mfcc_features*2] = mfcc_vars
                recording_features[j, 4+num_mfcc_features*2:4+num_mfcc_features*3] = mfcc_skew

    recording_features = recording_features.flatten()

    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant], recording_features))

    return np.asarray(features, dtype=np.float32)


# Extract features from the data.
def get_each_feature(data, recordings):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    # Extract recording locations and data. Identify when a location is present,
    # and compute the mean, variance, and skewness of each recording. If there
    # are multiple recordings for one location, then extract features from the
    # last recording.
    locations = get_locations(data)


    num_mfcc_features = 10
    recording_features = np.zeros((4 + num_mfcc_features * 3,), dtype=float)

    if np.size(recordings)>0:
        # compute mfcc feature
        mfcc = librosa.feature.mfcc(y=recordings.astype('float'), n_mfcc=num_mfcc_features)
        mfcc -= (np.mean(mfcc, axis=0) + 1e-8) # normalization

        mfcc_means = mfcc.mean(axis=1)
        mfcc_vars = mfcc.var(axis=1)
        mfcc_skew = sp.stats.skew(mfcc, axis=1)

        recording_features[0] = 1
        recording_features[1] = np.mean(recordings)
        recording_features[2] = np.var(recordings)
        recording_features[3] = sp.stats.skew(recordings)
        recording_features[4:4+num_mfcc_features] = mfcc_means
        recording_features[4+num_mfcc_features:4+num_mfcc_features*2] = mfcc_vars
        recording_features[4+num_mfcc_features*2:4+num_mfcc_features*3] = mfcc_skew

    recording_features = recording_features.flatten()

    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant], recording_features))

    return np.asarray(features, dtype=np.float32)
